"""
This is the C++ implementation of pylimer_tools. 

The subpackage (actually compiled code) :module:`pylimer_tools_cpp.pylimer_tools_cpp` 
should be imported automatically when you `import pylimer_tools_cpp`, 
meaning there is no need to use the double `pylimer_tools_cpp.pylimer_tools_cpp`,
a single one is sufficient.
"""

from .pylimer_tools_cpp import *
